/*
This file is part of myFitter, a C++ library for maximum likelihood fits.
Copyright (C) 2012  Martin Wiebusch

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include "fitter.hpp"

#include <gsl/gsl_vector.h>
#include <gsl/gsl_blas.h>

#include <boost/numeric/ublas/io.hpp>


namespace myfitter {

// parameter mappings
//
// pmap maps [-inf,+inf] to [0,+inv] with the asymptotic behaviour
// pmap(x) = x for x -> +inf. dpmap is the derivative of pmap and
// invpmap is the inverse function

static inline double pmap(double x) { return 0.5*(x + std::sqrt(x*x + 1.)); }

static inline double dpmap(double x)
{ return 0.5*(1. + x/std::sqrt(x*x + 1.)); }

static inline double invpmap(double x) { return x - 0.25/x; }


// gsl minimization functions

static double fcn(const gsl_vector * x, void * params)
{
    Fitter::FCNParameters *fcnparams =
        static_cast<Fitter::FCNParameters *>(params);
    Model* model = fcnparams->model;
    Fitter* fitter = fcnparams->fitter;
    const ObservableVector* cvals = fcnparams->cvals;
    const IndexVector& freepar = fcnparams->freepar;

    if(fcnparams->status) return 0.;
 
    for(int i=0; i<int(freepar.size()); ++i) {
        double pari;
        if(!model->has_lower_limit(freepar[i]) &&
           !model->has_upper_limit(freepar[i])) {
        
            pari = gsl_vector_get(x, i)*model->scale(freepar[i]);

        } else if(model->has_lower_limit(freepar[i]) &&
                  !model->has_upper_limit(freepar[i])) {

            double xi = gsl_vector_get(x, i);
            pari = model->lower_limit(freepar[i]) + 
                pmap(xi)*model->scale(freepar[i]);

        } else if(!model->has_lower_limit(freepar[i]) &&
                  model->has_upper_limit(freepar[i])) {

            double xi = gsl_vector_get(x, i);
            pari = model->upper_limit(freepar[i]) -
                pmap(-xi)*model->scale(freepar[i]);
            
        } else {

            double xi = gsl_vector_get(x, i);
            double delta = (model->upper_limit(freepar[i]) -
                            model->lower_limit(freepar[i]))/
                model->scale(freepar[i]);
            double q = invpmap(delta) - pmap(xi);
            pari = model->upper_limit(freepar[i]) - 
                pmap(q)*model->scale(freepar[i]); 
        }

        model->parameter(freepar[i], pari);
    }
    if(fcnparams->verbose >= 2)
        std::cout << "parameters: " << model->parameters() << std::endl;
    if((fcnparams->status = model->calc())) return 0.;
    if(fcnparams->verbose >= 2)
        std::cout << "observables: " << model->observables() << std::endl;
    return fitter->likelihood_function().calc(model->observables(), *cvals);
}

static void fcndfcn(const gsl_vector * x, void * params,
                    double * f, gsl_vector * g)
{
    Fitter::FCNParameters *fcnparams =
        static_cast<Fitter::FCNParameters *>(params);
    Model* model = fcnparams->model;
    Fitter* fitter = fcnparams->fitter;
    const ObservableVector* cvals = fcnparams->cvals;
    const IndexVector& freepar = fcnparams->freepar;
    const IndexVector& claimedobs
        = fcnparams->fitter->likelihood_function().claimed_observables();

    if(fcnparams->status) return;

    *f = fcn(x, params);
    if(fcnparams->status) return;
    fcnparams->status = model->calc_deriv();
    if(fcnparams->verbose >= 2)
        std::cout << "derivatives: " << model->derivatives() << std::endl;
    if(fcnparams->status) return;
    ObservableVector dummy(model->nobservables());
    fitter->likelihood_function().calc_deriv(model->observables(),
                                             *cvals, dummy);

    ObservableVector gradlfunc(claimedobs.size());
    Matrix derivs(freepar.size(),claimedobs.size());
    for(int iobs=0; iobs<int(claimedobs.size()); ++iobs) {
        gradlfunc[iobs] = dummy[claimedobs[iobs]];
        for(int ipar=0; ipar < int(freepar.size()); ++ipar)
            derivs(ipar,iobs)
                = model->derivatives()(freepar[ipar],claimedobs[iobs]);
    }
    ParameterVector gradchisq(prod(derivs, gradlfunc));
    
    for(int i=0; i<int(freepar.size()); ++i) {
        if(!model->has_lower_limit(freepar[i]) &&
           !model->has_upper_limit(freepar[i])) {

            gsl_vector_set(g, i, gradchisq[i]*
                           model->scale(freepar[i]));

        } else if(model->has_lower_limit(freepar[i]) &&
                  !model->has_upper_limit(freepar[i])) {

            double xi = gsl_vector_get(x, i);
            gsl_vector_set(g, i, gradchisq[i]*
                           dpmap(xi)*model->scale(freepar[i]));

        } else if(!model->has_lower_limit(freepar[i]) &&
                  model->has_upper_limit(freepar[i])) {

            double xi = gsl_vector_get(x, i);
            gsl_vector_set(g, i, gradchisq[i]*
                           dpmap(-xi)*model->scale(freepar[i]));
            
        } else {

            double xi = gsl_vector_get(x, i);
            double delta = (model->upper_limit(freepar[i]) -
                            model->lower_limit(freepar[i]))/
                model->scale(freepar[i]);
            double q = delta - 0.25/delta - pmap(xi);
            gsl_vector_set(g, i, gradchisq[i]*
                           dpmap(q)*dpmap(xi)*model->scale(freepar[i]));
        }
    }
}

static void dfcn(const gsl_vector * x, void * params, gsl_vector * g)
{
    double chisq;
    fcndfcn(x, params, &chisq, g);
}


// class Fitter

Fitter::Fitter(int no)
    : nobs_(no), lfunc_(no),
      orthogonalization_tolerance_(1.e-6),
      minimizer_(gsl_multimin_fdfminimizer_conjugate_fr),
      minimizer_verbose_(0),
      minimizer_step_width_(0.01),
      minimizer_tolerance_(0.1),
      minimizer_precision_(1.e-3),
      minimizer_iterations_(200),
      minimizer_snap_distance_(0.1),
      inner_region_fraction_(0.1),
      inner_region_power_(1.),
      dvegas_verbose_(0), dvegas_nbins_(50),
      dvegas_nfirstshots_(1000),
      dvegas_nshots_(500),
      dvegas_niterations_(20),
      dvegas_precision_(0.01),
      pvalue_(Model::nan()),
      pvalue_error_(Model::nan()),
      reached_precision_goal_(false),
      failed_shot_ratio_(0.)
{ }

int Fitter::global_fit(Model& model, const ObservableVector& cvals)
{
    if(model.nsamples() <= 0) return local_fit(model);
    int besti = 0;
    double chi2min = lfunc_.calc(model.sample_observables(0), cvals);
    double chi2 = 0.;
    for(int i=1; i<model.nsamples(); ++i) {
        chi2 = lfunc_.calc(model.sample_observables(i), cvals);
        if(chi2 < chi2min) {
            chi2min = chi2;
            besti = i;
        }
    }
    model.parameters(model.sample_parameters(besti));
    return local_fit(model, cvals);
}

int Fitter::local_fit(Model& model, const ObservableVector& cvals)
{
    std::vector<bool> fix(model.nparameters());
    for(int i=0; i<model.nparameters(); ++i)
        fix[i] = model.fixed(i);

    int result = local_fit_(model, cvals);

    for(int i=0; i<model.nparameters(); ++i)
        if(!fix[i]) model.release(i);

    return result;
}

int Fitter::local_fit_(Model& model, const ObservableVector& cvals)
{
    FCNParameters params;
    params.status = 0;
    params.verbose = minimizer_verbose_;
    params.fitter = this;
    params.model = &model;
    params.cvals = &cvals;
    for(int i=0; i<model.nparameters(); ++i)
        if(!model.fixed(i)) params.freepar.push_back(i);
    int npar=params.freepar.size();
    if(npar == 0) {
        calc(model, cvals);
        return 0;
    }

    gsl_vector *x;
    x = gsl_vector_alloc(npar);
    double xi=0.;
    for(int i=0; i<npar; ++i) {
        if(!model.has_lower_limit(params.freepar[i]) &&
           !model.has_upper_limit(params.freepar[i])) {

            xi = model.parameter(params.freepar[i])/
                model.scale(params.freepar[i]);

        } else if(model.has_lower_limit(params.freepar[i]) &&
                  !model.has_upper_limit(params.freepar[i])) {

            xi = (model.parameter(params.freepar[i])
                  -model.lower_limit(params.freepar[i]))/
                model.scale(params.freepar[i]);
            xi = invpmap(xi);

        } else if(!model.has_lower_limit(params.freepar[i]) &&
                  model.has_upper_limit(params.freepar[i])) {

            xi = (model.upper_limit(params.freepar[i])-
                  model.parameter(params.freepar[i]))/
                model.scale(params.freepar[i]);
            xi = -invpmap(xi);
                
        } else if(model.has_lower_limit(params.freepar[i]) &&
                  model.has_upper_limit(params.freepar[i])) {

            double delta = (model.upper_limit(params.freepar[i]) -
                            model.lower_limit(params.freepar[i]))/
                model.scale(params.freepar[i]);
            double q = (model.upper_limit(params.freepar[i]) -
                        model.parameter(params.freepar[i]))/
                model.scale(params.freepar[i]);
            q = invpmap(q);
            xi = invpmap(invpmap(delta) - q);

        }
        gsl_vector_set(x, i, xi);
    }

    gsl_multimin_function_fdf func;     
    func.n = npar;
    func.f = fcn;
    func.df = dfcn;
    func.fdf = fcndfcn;
    func.params = &params;

    gsl_multimin_fdfminimizer *s;
    s = gsl_multimin_fdfminimizer_alloc(minimizer(), npar);
    gsl_multimin_fdfminimizer_set(s, &func, x,
                                  minimizer_step_width(),
                                  minimizer_tolerance());

    int iter = 0;
    int status;
    std::vector<double> dupper(model.nparameters(), -1.);
    std::vector<double> dlower(model.nparameters(), -1.);
    double d;
    double chisq0 = likelihood_function().calc(cvals,cvals);
    do {
        // check for snapping conditions
        for(int i=0; i<model.nparameters(); ++i) {
            if(model.fixed(i)) continue;
            if(model.has_upper_limit(i)) {
                d = std::abs(model.parameter(i)-model.upper_limit(i));
                if(d < minimizer_snap_distance()*model.scale(i) &&
                   d < dupper[i]) {

                    model.parameter(i, model.upper_limit(i));
                    model.fix(i);
                    gsl_multimin_fdfminimizer_free(s);
                    gsl_vector_free(x);
                    if(minimizer_verbose_ >= 1)
                        std::cout << "snapping parameter " << i
                                  << " to " << model.upper_limit(i)
                                  << std::endl;
                    return local_fit_(model, cvals);
                    
                } else
                    dupper[i] = d;
            }
            if(model.has_lower_limit(i)) {
                d = std::abs(model.parameter(i)-model.lower_limit(i));
                if(d < minimizer_snap_distance()*model.scale(i) &&
                   d < dlower[i]) {

                    model.parameter(i, model.lower_limit(i));
                    model.fix(i);
                    gsl_multimin_fdfminimizer_free(s);
                    gsl_vector_free(x);
                    if(minimizer_verbose_ >= 1)
                        std::cout << "snapping parameter " << i
                                  << " to " << model.lower_limit(i)
                                  << std::endl;
                    return local_fit_(model, cvals);

                } else
                    dlower[i] = d;
            }
        }

        status = gsl_multimin_fdfminimizer_iterate(s);  
        if(params.status) return params.status;
        if(status) return status;
        if(minimizer_verbose_ >= 1)
            std::cout << "chisquare: " << (s->f - chisq0) << std::endl;

        status = gsl_multimin_test_gradient(s->gradient,
                                            minimizer_precision());
        if(status == GSL_SUCCESS) break;
        iter++;
    } while (status == GSL_CONTINUE && iter < minimizer_iterations());

     
    model.chisquare(fcn(s->x, &params) - chisq0);

    gsl_multimin_fdfminimizer_free(s);
    gsl_vector_free(x);

    if(status != GSL_SUCCESS) return GSL_EMAXITER;
    return 0;
}


}
