#include <iostream>
#include <boost/numeric/ublas/io.hpp>

#include "fitter.hpp"
#include "likelihoodcomponents.hpp"

#include "comparison.hpp"

using namespace myfitter;


inline bool differ(double a, double b) { return differ(a, b, 1.e-4, 1.e-10); }


class LinearModel : public Model {
public:
    enum {P1, P2, NPAR};
    enum {O1, O2, O3, O4, NOBS};

    LinearModel() : Model(NPAR,NOBS) { }
    LinearModel(const LinearModel& m) : Model(m) { }

    virtual int calc() {
        observables_[O1] = parameter(P1) + 1.*parameter(P2);
        observables_[O2] = parameter(P1) + 2.*parameter(P2);
        observables_[O3] = parameter(P1) + 3.*parameter(P2);
        observables_[O4] = parameter(P1) + 4.*parameter(P2);
        needs_update_ = false;
        return 0;
    }
};

int main()
{
    std::cout.precision(15);
    typedef LinearModel Mod;

    Fitter fitter(Mod::NOBS);
    fitter.minimizer_precision(1.e-5);

    LinearModel model;
    model.derivative_epsilon(1.e-7);
    model.scale(Mod::P1, 0.1);
    model.scale(Mod::P2, 0.1);
    model.parameter(Mod::P1, 2.2);
    model.parameter(Mod::P2, 0.7);
    model.scan_range(Mod::P1, 0., 5.);
    model.scan_range(Mod::P2, 0., 2.);

    fitter.likelihood_function().add(GaussianLC(Mod::O1, 3., 1.2));
    fitter.likelihood_function().add(GaussianLC(Mod::O2, 3.5, 1.));
    fitter.likelihood_function().add(GaussianLC(Mod::O3, 4.2, 0.4));
    fitter.likelihood_function().add(GaussianLC(Mod::O4, 5.1, 0.3));

    ObservableVector meas(Mod::NOBS);
    fitter.likelihood_function().get_central_values(meas);
    model.calc();
    double chi2 = fitter.likelihood_function().calc(model.observables(),meas) -
        fitter.likelihood_function().calc(meas,meas);
    if(differ(chi2, 0.190556)) return 1;

    model.clear();
    model.scan(100);
    if(fitter.global_fit(model)) return 1;
    if(differ(model.chisquare(), 0.111204)) return 1;
    if(differ(model.parameter(Mod::P1), 1.94983)) return 1;
    if(differ(model.parameter(Mod::P2), 0.779264)) return 1;
    if(differ(model.observables()[Mod::O1], 2.72909)) return 1;
    if(differ(model.observables()[Mod::O2], 3.50836)) return 1;
    if(differ(model.observables()[Mod::O3], 4.28762)) return 1;
    if(differ(model.observables()[Mod::O4], 5.06689)) return 1;

    model.parameter(Mod::P1, 2.);
    model.fix(Mod::P1);
    model.clear();
    model.scan(100);
    if(fitter.global_fit(model)) return 1;
    if(differ(model.chisquare(), 0.113393)) return 1;
    if(differ(model.parameter(Mod::P1), 2.)) return 1;
    if(differ(model.parameter(Mod::P2), 0.76541)) return 1;
    if(differ(model.observables()[Mod::O1], 2.76541)) return 1;
    if(differ(model.observables()[Mod::O2], 3.53082)) return 1;
    if(differ(model.observables()[Mod::O3], 4.29623)) return 1;
    if(differ(model.observables()[Mod::O4], 5.06164)) return 1;

    model.release(Mod::P1);
    model.lower_limit(Mod::P1, 1.9);
    model.clear();
    model.scan(100);
    if(fitter.global_fit(model)) return 1;
    if(differ(model.chisquare(), 0.111204)) return 1;
    if(differ(model.parameter(Mod::P1), 1.94983)) return 1;
    if(differ(model.parameter(Mod::P2), 0.779264)) return 1;
    if(differ(model.observables()[Mod::O1], 2.72909)) return 1;
    if(differ(model.observables()[Mod::O2], 3.50836)) return 1;
    if(differ(model.observables()[Mod::O3], 4.28762)) return 1;
    if(differ(model.observables()[Mod::O4], 5.06689)) return 1;

    model.unset_lower_limit(Mod::P1);
    model.upper_limit(Mod::P1, 2.0);
    model.clear();
    model.scan(100);
    if(fitter.global_fit(model)) return 1;
    if(differ(model.chisquare(), 0.111204)) return 1;
    if(differ(model.parameter(Mod::P1), 1.94983)) return 1;
    if(differ(model.parameter(Mod::P2), 0.779264)) return 1;
    if(differ(model.observables()[Mod::O1], 2.72909)) return 1;
    if(differ(model.observables()[Mod::O2], 3.50836)) return 1;
    if(differ(model.observables()[Mod::O3], 4.28762)) return 1;
    if(differ(model.observables()[Mod::O4], 5.06689)) return 1;

    model.range(Mod::P1, 1.8, 2.1);
    model.clear();
    model.scan(100);
    if(fitter.global_fit(model)) return 1;
    if(differ(model.chisquare(), 0.111204)) return 1;
    if(differ(model.parameter(Mod::P1), 1.94983)) return 1;
    if(differ(model.parameter(Mod::P2), 0.779264)) return 1;
    if(differ(model.observables()[Mod::O1], 2.72909)) return 1;
    if(differ(model.observables()[Mod::O2], 3.50836)) return 1;
    if(differ(model.observables()[Mod::O3], 4.28762)) return 1;
    if(differ(model.observables()[Mod::O4], 5.06689)) return 1;

    model.range(Mod::P1, 2.0, Model::nan());
    model.clear();
    model.scan(100);
    if(fitter.global_fit(model)) return 1;
    if(differ(model.chisquare(), 0.113393)) return 1;
    if(differ(model.parameter(Mod::P1), 2.)) return 1;
    if(differ(model.parameter(Mod::P2), 0.76541)) return 1;
    if(differ(model.observables()[Mod::O1], 2.76541)) return 1;
    if(differ(model.observables()[Mod::O2], 3.53082)) return 1;
    if(differ(model.observables()[Mod::O3], 4.29623)) return 1;
    if(differ(model.observables()[Mod::O4], 5.06164)) return 1;

    return 0;
}
