#include <iostream>

#include "fitter.hpp"
#include "likelihoodcomponents.hpp"
#include "comparison.hpp"

#include <boost/numeric/ublas/io.hpp>

using namespace myfitter;

class LinearModel : public Model {
public:
    Matrix mat;
    ObservableVector vec;

    LinearModel(int npar, int nobs) :
        Model(npar,nobs), mat(nobs,npar), vec(nobs) { }
    LinearModel(const LinearModel& m) : Model(m), mat(m.mat), vec(m.vec) { }
    virtual const LinearModel& operator=(const LinearModel& m) {
        Model::operator=(m);
        mat = m.mat;
        vec = m.vec;
        return *this;
    }

    virtual int calc() {
        observables_
            .assign(boost::numeric::ublas::prod(mat, parameters()) + vec);
        needs_update_ = false;
        return 0;
    }
};

int main()
{
    typedef LinearModel Mod;


    // initialize model

    int nobs = 3;
    int npar1 = 1;
    int npar2 = 2;

    LinearModel model1(npar1,nobs);
    model1.derivative_epsilon(1.e-7);

    LinearModel model2(npar2,nobs);
    model2.derivative_epsilon(1.e-7);

    model1.scale(0, 0.1);
    model1.scan_range(0, -10., +10.);

    model2.scale(0, 0.1);
    model2.scale(1, 0.1);
    model2.scan_range(0, -10., +10.);
    model2.scan_range(1, -10., +10.);

    for(int i=0; i<nobs; ++i) {
        model1.vec(i) = 0.;
        model2.vec(i) = 0.;
        for(int j=0; j<npar1; ++j)
            model1.mat(i,j) = 0.;
        for(int j=0; j<npar2; ++j)
            model2.mat(i,j) = 0.;
    }

    model1.mat(0, 0) = 1.;
    model1.mat(2, 0) = 1.;

    model2.mat(0, 0) = 1.;
    model2.mat(1, 1) = 1.;


    // initialise fitter and likelihood function

    Fitter fitter(nobs);
    fitter.dvegas_precision(0.02);
    fitter.dvegas_nfirstshots(2000);
    fitter.dvegas_nshots(1000);
    // fitter.dvegas_verbosity(2);

    fitter.likelihood_function().add(GaussianLC(0, 1.2, 1.));
    fitter.likelihood_function().add(GaussianLC(1, 0.3, 1.));
    fitter.likelihood_function().add(GaussianLC(2, 1.0, 1.));

    
    // scan and fit models

    model1.clear();
    model1.scan(1000);
    if(fitter.global_fit(model1)) return 1;
    // std::cout << "\nmodel 1 parameters: " << model1.parameter(0)
    //           << "\nmodel 1 chi^2:      " << model1.chisquare()
    //           << std::endl;

    model2.clear();
    model2.scan(1000);
    if(fitter.global_fit(model2)) return 1;
    // std::cout << "\nmodel 2 parameters: "
    //           << model2.parameter(0) << " " << model2.parameter(1) 
    //           << "\nmodel 2 chi^2:      " << model2.chisquare()
    //           << std::endl;


    // calculate relative p-value

    fitter.calc_lrt_pvalue(model1, model2);
    if(!fitter.reached_precision_goal()) return 1;
    if(fitter.failed_shot_ratio() > 0.005) return 1;
    if(differ(fitter.pvalue(), 0.092, 0.04, 1.e-10)) return 1;

    // std::cout << "\nprecision goal reached: ";
    // if(fitter.reached_precision_goal())
    //     std::cout << "yes";
    // else
    //     std::cout << "no";
    // std::cout << "\nfailed shot ratio:      " << fitter.failed_shot_ratio();
    // std::cout << "\np-value:                " << fitter.pvalue();
    // std::cout << "\np-value error:          " << fitter.pvalue_error()
    //           << std::endl;


    return 0;
}
